extern crate lodepng;

use std::collections::HashMap;
use std::iter::Iterator;
use std::path::Path;
use vector::*;
use std::io;
use std;
use lwofront;
use ultraback;
use math;

type Image = lodepng::Bitmap<lodepng::RGBA<u8>>;
type Vector3f = Vector3<f32>;

struct MergeResult {
	vmap: Box<lwofront::Vmap>,
	pnts: Box<lwofront::Pnts>,
	pols: Box<lwofront::Pols>
}

struct Pts<'a> {
	point: &'a lwofront::Point,
	normal: Vector3f,
	tc_img: Option<(&'a Vec<f32>, &'a Image)>
}

fn rgba8888_to_rgba5551( color: &lodepng::RGBA<u8> ) -> u16 {
	((color.r as u16 & 0xF8) << 8)
	| ((color.g as u16 & 0xF8) << 3)
	| ((color.b as u16 & 0xF8) >> 2)
	| (color.a as u16 >> 7)
}

fn calculate_normal(vi: usize, points: &lwofront::Pnts, polygons: &lwofront::Pols) -> Vector3f {
	let polys = polygons.v.iter()
		.filter( |&&(_,ref face)| {
			face.iter().find( |vidx| {
				**vidx == vi as u32
			} )
			.is_some()
		} )
		.map( |&(_,ref face)| face );				// turning map into iterator?

	let mut acc_normal = Vector3f { v: [0.0, 0.0, 0.0] };

	let mut i = 0;
	for (_,polygon) in polys.enumerate() {				// return value of the loop?
		let (p0,p1,p2) = (
			&points.v[polygon[0] as usize],
			&points.v[polygon[1] as usize],
			&points.v[polygon[2] as usize]
		);

		let u = math::sub_vec(&p1, &p0);
		let v = math::sub_vec(&p2, &p0);
		acc_normal = math::add_vec(
			&acc_normal,
			&math::normalize_vector( &math::cross_product( &u, &v ) )
		);

		i += 1;
	}

	math::scale_uniformly_vec( &acc_normal, (1.0 / i as f32) )
}

fn merge_vmap_vmad(vmap: &lwofront::Vmap, vmad: &lwofront::Vmad, pols: &lwofront::Pols, pnts: &lwofront::Pnts) -> MergeResult {
	let mut result = MergeResult {
		vmap: Box::new( vmap.clone() ),
		pnts: Box::new( pnts.clone() ),
		pols: Box::new( pols.clone() )
	};

	for (vi, vertex) in pnts.v.iter().enumerate() {
		let vmad_v = vmad.v.iter()
			.filter( |vi_pi_val| {
				vi_pi_val.vi == vi as u32
			} );

		for vi_pi_val in vmad_v {
			let new_vi = result.pnts.v.len() as u32;
			result.pnts.v.push( vertex.clone() );
			result.vmap.v.push(
				lwofront::VmapViVal {
					vi: new_vi,
					val: vi_pi_val.val.clone()
				}
			);

			let pi = vi_pi_val.pi as usize;
			let pvi = result.pols.v[pi].1.iter().enumerate()
				.find( |&(_,vidx)| {
					*vidx == vi as u32
				} )
				.unwrap().0 as usize;
			result.pols.v[pi].1[pvi] = new_vi;
		}
	}

	result

}

fn convert_polygon(polygon: &lwofront::Polygon) -> ultraback::Face {
	match polygon.len() {
		3 => ultraback::Face::Tri {
			ids: Vector3::<i16> {
				v: [
					polygon[0] as i16,		// TODO: check that we can actualy cast without loss
					polygon[2] as i16,
					polygon[1] as i16
				]
			}
		},
		4 => ultraback::Face::Quad {
			ids: Vector4::<i16> {
				v: [
					polygon[1] as i16,
					polygon[0] as i16,
					polygon[3] as i16,
					polygon[2] as i16
				]
			}
		},
		_ => panic!("Unsupported polygon"),			// TODO: don't panic, tessalate!
	}
}

fn convert_point(pnt: &Pts) -> ultraback::Vertex {
	let texcoord_scale = 2.0*32.0;	// 32 for shifting 5 bits left, 2 for additional shift; TODO: this should be set by the user
	let point = pnt.point;
	let normal = math::scale_uniformly_vec(&pnt.normal, 127.0);
	
	for &coord in point.v.iter() {
		if ( coord < i16::min_value() as f32 ) || ( coord > i16::max_value() as f32 ) {
			println!("Warning: vertex coordinate doesn't fit in 16 bits")
		}
	}

	let tc = match pnt.tc_img {
		Some(tc_img) => {
			let mut tc = tc_img.0.clone();
			tc[0] *= tc_img.1.width as f32;
			tc[1] *= tc_img.1.height as f32;

			if (tc[0].abs() as i16 > std::i16::MAX/(texcoord_scale as i16))
			|| (tc[1].abs() as i16 > std::i16::MAX/(texcoord_scale as i16)) {
				println!("Warning: texture coordinate out of range. Setting to 0.");
				tc[0] = 0.0;
				tc[1] = 0.0;
			}

			tc
		},
		None => vec![0.0, 0.0]
	};

	ultraback::Vertex {
		pos: Vector3::<i16> {
			v: [
				point.v[0] as i16,
				point.v[1] as i16,
				-point.v[2] as i16	// transform to RHS
			]
		},
		tc: Vector2::<i16> {
			v: [
				(texcoord_scale * tc[0]) as i16,
			   	(-texcoord_scale * tc[1]) as i16	// transform texture coordsys
			]
		},
		col: Vector4::<u8> {
			v: [
				normal.v[0] as u8,
				normal.v[1] as u8,
				normal.v[2] as u8,
				0xFF
			]
		}
	}
}

fn convert_layer(src_layer: &lwofront::Layr, &(_, surface): &(usize, &lwofront::Surf), textures: &Vec<(usize, Option<Image>)>) -> ultraback::Layer {
	let mut dst_layer = ultraback::Layer {
		verts: Vec::new(),
		faces: Vec::new(),
		face2mtl: Vec::new()
	};

	for (i,_) in src_layer.polygons.v.iter().enumerate() {
		dst_layer.face2mtl.push(
			(i, (i,surface).0)
		);
	}

	// find TXUV type VMAP and VMAD
	let vmad = src_layer.vmads.iter()
		.find( |vmad| {
			match vmad.typecode {
				lwofront::VmadType::TXUV => true,
				_ => false
			}
		} );

	let (vmap, pnts, pols) =
		match src_layer.vmaps.iter()
			.find( |vmap| {
				match vmap.typecode {
					lwofront::VmapType::TXUV => true,
					_ => false
				}
			} )
		{
			None => (None, src_layer.points.clone(), src_layer.polygons.clone()),
			Some(vmap) => {
				let merged = match vmad {
					Some(vmad) => {
						merge_vmap_vmad(
							&vmap,
							&vmad,
							&src_layer.polygons,
							&src_layer.points
						)
					},
					None => {
						MergeResult {
							vmap: vmap.clone(),
							pnts: src_layer.points.clone(),
							pols: src_layer.polygons.clone()
						}
					}
				};

				(Some(merged.vmap), merged.pnts, merged.pols)
			}
		};

	// find tex index
	let get_texid = |blok_subchunks: &Vec<lwofront::BlokSubchunk>| -> Option<lwofront::Vx> {
		blok_subchunks.iter().find( |&bs| {
			match bs {
				&lwofront::BlokSubchunk::Imag { .. } => true,
				_ => false
			}
		} ).and_then( |bs| {
			match bs {
				&lwofront::BlokSubchunk::Imag { ref texture } => Some(*texture - 1),	// -1 !!!!!
				_ => None
			}
		} )
	};

	let mut pts: Vec<Pts> = Vec::new();
	for (vi, vertex) in pnts.v.iter().enumerate() {
		// find texcoords
		let tc_img = match vmap {
			Some(ref vm) => {
				let tc = vm.v.iter()
					.find( |&vi_val| {
						vi_val.vi == vi as u32
					} )
					.and_then( |ref vi_val| {
						Some(&vi_val.val)
					} ).unwrap();

				let texi = surface.attributes.iter().find( |&attr| {
					match attr {
						&lwofront::SurfSubchunk::Blok { ref header, .. } => {
							match header {
								&lwofront::BlokHeader::Imap { .. } => true,
								_ => false
							}
						},
						_ => false
					}
				} ).and_then( |attr| {
					match attr {
						&lwofront::SurfSubchunk::Blok { ref header, ref attributes } => {
							match header {
								&lwofront::BlokHeader::Imap { .. } => get_texid( attributes ), 	
								_ => None
							}
						},
						_ => None
					}
				} );

				match texi {
					Some(ti) if ti < textures.len() as u32 => {
						match textures[ti as usize].1.as_ref() {
							Some(tex) => Some((tc, tex)),
							None => None
						}
					},
					_ => None
				}

			},
			None => None
		};

		pts.push( Pts {
			point: vertex,
			normal: calculate_normal( vi, &pnts, &pols ),
			tc_img: tc_img
		} );
	}

	// TODO: split vertex if polygons have different materials

	for pnt in pts.iter() {
		dst_layer.verts.push( convert_point( &pnt ) );
	}

	for polygon in &pols.v {
		dst_layer.faces.push( convert_polygon( &polygon.1 ) );
	}

	dst_layer
}

fn convert_surface(src_surface: &lwofront::Surf) -> ultraback::Material {
	let mut dst_mtl = ultraback::Material {
		tex_id: None,
		color: Vector3::<f32> { v: [0.5, 0.5, 0.5] }
	};

	// assuming IMAP context
	let handle_blok_attributes = |attributes, mtl: &mut ultraback::Material| {
		for attribute in attributes {
			match attribute {
				&lwofront::BlokSubchunk::Imag { texture } => {
					mtl.tex_id = Some((texture-1)  as usize);	// -1 !!!!!!
				},
				/*
				&lwofront::BlokSubchunk::Tmap { ref attributes } => {
					for attribute in attributes {
						match attribute {
							&lwofront::TmapSubchunk::Size { vector: ref v, .. } => {
								// TODO
							},
							_ => ()
						}
					}
				},
				&lwofront::BlokSubchunk::Axis { ref texture_axis } => {
					// TODO
				},
				*/
				_ => ()
			}
		}
	};

	for attribute in &src_surface.attributes {
		match attribute {
			&lwofront::SurfSubchunk::Colr { ref base, .. } => {
				dst_mtl.color = Vector3::<f32> { v: base.v };
			},
			&lwofront::SurfSubchunk::Blok { ref header, ref attributes } => {
				match header {
					&lwofront::BlokHeader::Imap {..} => {
						handle_blok_attributes( attributes, &mut dst_mtl );
					},
					_ => ()
				}
			},
			_ => ()
		}
	}

	dst_mtl
}

fn open_image(img_path: &str) -> Result<Image, io::Error> {
	let mut image = try!(
		lodepng::decode32_file( img_path )
		.map_err( |_| io::Error::new(
			io::ErrorKind::Other,
			format!( "Couldn't open image" )
		) )
	);

	// ignore alpha channel for the moment; TODO: this should depend on option passed by user
	for i in 0.. {
		match image.buffer.get_mut(i) {
			Some(c) => c.a = 0xFF,
			None => break
		};
	}

	return Ok(image)
}

fn convert_image(bitmap: &Image) -> (ultraback::Texture, Option<&'static str>) {
	let mut texture = ultraback::Texture {
		width: bitmap.width as u16,
		height: bitmap.height as u16,
		data: ultraback::TextureData::Unknown { pixels: Vec::new() },
	};

	let mut histogram = HashMap::new();				// poor performance!
	let buffer = bitmap.buffer.as_ref();
	for i in 0..buffer.len() {
		*histogram.entry( buffer[i] ).or_insert( 0 ) += 1;
	}

	// palette
	if histogram.keys().len() <= 16 {
		texture.data = ultraback::TextureData::CI4 {
			pixels: Vec::new(),
			palette: histogram.keys().map( rgba8888_to_rgba5551 ).collect()
		};
	} else {
		texture.data = ultraback::TextureData::RGBA16 { pixels: Vec::new() };
	}

	// convert pixels
	let mut warning = None;
	match texture.data {
		ultraback::TextureData::CI4 { ref mut pixels, ref palette } => {
			for i in 0.. {
				let colorl = match bitmap.buffer.get( 2*i ) {
					Some(c) => {
						palette.iter().enumerate()
						.find( |&(_,col)| *col == rgba8888_to_rgba5551( c ))
						.unwrap().0 as u8
					},
					None => break
				};
				let colorr = match bitmap.buffer.get( 2*i + 1 ) {
					Some(c) => {
						palette.iter().enumerate()
						.find( |&(_,col)| *col == rgba8888_to_rgba5551( c ))
						.unwrap().0 as u8
					},
					None => 0
				};
				pixels.push( (colorl << 4) | colorr );
			}

			if pixels.len() > 4096 {
				warning = Some("texture won't fit in TMEM");
			}
		},
		ultraback::TextureData::RGBA16 { ref mut pixels } => {
			for i in 0.. {
				pixels.push( match bitmap.buffer.get( i ) {
					Some(c) => rgba8888_to_rgba5551( c ),
					None => break
				});
			}

			if pixels.len() > 1024 {
				warning = Some("texture won't fit in TMEM");
			}
		},
		ultraback::TextureData::Unknown { ref mut pixels } => {
			for i in 0.. {
				match bitmap.buffer.get( i ) {
					Some(c) => {
						pixels.push( c.r );
						pixels.push( c.g );
						pixels.push( c.b );
						pixels.push( c.a );
					},
					None => break
				}
			}

			warning = Some("Unknown texture format");
		}
	}

	(texture, warning)
}

pub fn convert(src_mesh: &lwofront::Object, src_mesh_dir: &str) -> Result<Box<ultraback::Mesh>, io::Error> {
	let mut dst_mesh = Box::new( ultraback::Mesh {
		layers: Vec::new(),
		materials: Vec::new(),
		textures: Vec::new()
	} );

	// textures
	let mut textures = Vec::new();
	for (idx, clip) in src_mesh.clips.iter().enumerate() {
		let mut imgpath = String::new();
		for subchunk in &clip.vsub {
			match subchunk {
				&lwofront::ClipSubchunk::Stil { ref name } => {
					imgpath = src_mesh_dir.to_owned()
					+ "/"
					+ Path::new( name )
						.file_name()
						.unwrap()
						.to_str()
						.unwrap()
				},

				_ => ()
			}
		};

		textures.push( (
			idx,
			match open_image( imgpath.as_str() ) {
				Ok(img) => {
					let converted = convert_image( &img );
					match converted.1 {
						Some(warning) => println!( "{}: {}", imgpath, warning ),
						None => ()
					}
					dst_mesh.textures.push( converted.0 );
					Some(img)
				},
				Err(_) => {
					println!("Warning: Couldn't open texture: {}", imgpath );

					dst_mesh.textures.push( ultraback::Texture {
						width: 0,
						height: 0,
						data: ultraback::TextureData::Unknown { pixels: Vec::new() },
					} );

					None
				}
			}
		) );
	}

	// surfaces
	for surface in &src_mesh.surfaces {
		dst_mesh.materials.push(
			convert_surface( surface )
		);
	}

	// layers
	for src_layer in &src_mesh.layers {
		// currently one surface per layer allowed
		let (i, surface) =
			src_mesh.surfaces.iter().enumerate()
			.find( |&(_,surf)| {
				src_layer.ptags.v.iter()
				.find( |&&(_,tagidx)| {
					surf.name == src_mesh.tags.v[tagidx as usize]
				} ).is_some()
			} ).unwrap();

		dst_mesh.layers.push(
			convert_layer(
				src_layer,
				&(i, surface),
				&textures
			)
		);
	}

	Ok(dst_mesh)
}

