use std::fs::File;
use std::result;
use std::io;
//use std::io::Read;
//use std::io::BufRead;
//use std::mem;
use std::error;
use std::fmt;
use stream;
use vector;

type Vector3f = vector::Vector3<f32>;
type Result<T> = result::Result<T, Box<error::Error>>;
type BufReader = io::BufReader<File>;
type Stream = stream::Stream<File>;
type Ordinal = Vec<u8>;
pub type Vx = u32;
pub type Polygon = Vec<u32>;
pub type Point = Vector3f;

#[derive(Debug)]
enum LwoError {
	//Io(io::Error),
	Invalid,
	UnknownBlokHeader,
}

impl fmt::Display for LwoError {
	fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
		match *self {
			//LwoError::Io(ref err) => write!(f, "IO error: {}", err),
			LwoError::Invalid => write!( f, "Invalid" ),
			LwoError::UnknownBlokHeader => write!( f, "Unknown BLOK header" ),
		}
	}
}

impl error::Error for LwoError {
	fn description(&self) -> &str {
		match *self {
			//LwoError::Io(ref err) => err.description(),
			LwoError::Invalid => "IO error",
			LwoError::UnknownBlokHeader => "Unknown BLOK header",
		}
	}
	
	fn cause(&self) -> Option<&error::Error> {
		match *self {
			//LwoError::Io(ref err) => Some(err),
			LwoError::Invalid => None,
			LwoError::UnknownBlokHeader => None,
		}
	}
}

enum Literals {
	LWO2 = 0x4c574f32,
	TXUV = 0x54585556,
	SURF = 0x53555246,
	FACE = 0x46414354,
}

enum ChunkIDs {
	FORM = 0x464f524d,
	LAYR = 0x4c415952,
	PNTS = 0x504e5453,
	VMAP = 0x564d4150,
	VMAD = 0x564d4144,
	POLS = 0x504f4c53,
	TAGS = 0x54414753,
	PTAG = 0x50544147,
	CLIP = 0x434c4950,
	SURF = 0x53555246,
}
enum SubchunkIDs {
	STIL = 0x5354494c,
	COLR = 0x434f4c52,
	DIFF = 0x44494646,
	SPEC = 0x53504543,
	GLOS = 0x474c4f53,
	BLOK = 0x424c4f4b,
	IMAP = 0x494d4150,
	CHAN = 0x4348414e,
	ENAB = 0x454e4142,
	OPAC = 0x4f504143,
	//AXIS = 0x
	//NEGA = 0x4e454741,
	TMAP = 0x544d4150,
	CNTR = 0x434e5452,
	SIZE = 0x53495a45,
	ROTA = 0x524f5441,
	FALL = 0x46414c4c,
	OREF = 0x4f524546,
	CSYS = 0x43535953,
	PROJ = 0x50524f4a,
	AXIS = 0x41584953,
	IMAG = 0x494d4147,
	WRAP = 0x57524150,
	WRPW = 0x57525057,
	WRPH = 0x57525048,
	VMAP = 0x564d4150,
	AAST = 0x41415354,
	PIXB = 0x50495842,
}

struct Chunk {
	tag: u32,
	length: u32,
}

pub struct Object {
	pub tags: Tags,
	pub layers: Vec<Box<Layr>>,
	pub clips: Vec<Clip>,
	pub surfaces: Vec<Box<Surf>>,
}

pub struct Layr {
	pub number: u16,
	pub flags: u16,
	pub pivot: Vector3f,
	pub name: String,
	pub parent: u16,

	/* data from chunks belonging to layer */
	pub points: Box<Pnts>,
	pub vmaps: Vec<Box<Vmap>>,
	pub vmads: Vec<Box<Vmad>>,
	pub polygons: Box<Pols>,
	pub ptags: Box<Ptag>
}

#[derive(Clone)]
pub struct Pnts {
	pub v: Vec<Point>,
}

#[derive(Clone)]
pub enum VmapType {
	TXUV,
	Unknown(u32)
}
#[derive(Clone)]
pub struct VmapViVal {
	pub vi: Vx,
	pub val: Vec<f32>
}
#[derive(Clone)]
pub struct Vmap {
	pub typecode: VmapType,
	pub dimension: u16,
	pub name: String,
	pub v: Vec<VmapViVal>,
}

pub enum VmadType {
	TXUV,
	Unknown(u32)
}
pub struct VmadViPiVal {
	pub vi: Vx,
	pub pi: Vx,
	pub val: Vec<f32>,
}
pub struct Vmad {
	pub typecode: VmadType,
	pub dimension: u16,
	pub name: String,
	pub v: Vec<VmadViPiVal>,
}

#[derive(Clone)]
pub enum PolsType {
	FACE,
	Unknown(u32)
}
#[derive(Clone)]
pub struct Pols {
	pub typecode: PolsType,
	pub v: Vec<(u16, Polygon)>,	// u16 contains flags
}
pub struct Tags {
	pub v: Vec<String>,
}

pub enum PtagType {
	SURF,
	Unknown(u32)
}
pub struct Ptag {
	pub typecode: PtagType,
	pub v: Vec<(u32, u16)>,
}
pub enum ClipSubchunk {
	Stil { name: String },		// name doesn't contain '/0' at the end
	None,
}
pub struct Clip {
	pub index: u32,
	pub vsub: Vec<ClipSubchunk>,
}
pub enum SurfSubchunk {
	Colr { base: Vector3f, envelope: u32 },
	Diff { intensity: f32, envelope: u32 },
	Spec { intensity: f32, envelope: u32 },
	Glos { intensity: f32, envelope: u32 },
	Blok { header: BlokHeader, attributes: Vec<BlokSubchunk> },
}
pub struct Surf {
	pub name: String,
	pub source: String,
	pub attributes: Vec<SurfSubchunk>,
}

struct Subchunk {
	tag: u32,
	length: u16,
}
pub enum BlokHeader {
	Imap { ordinal: Ordinal, attributes: Vec<BlokHeaderSubchunk> },
	Proc { ordinal: Ordinal, attributes: Vec<BlokHeaderSubchunk> },
	Grad { ordinal: Ordinal, attributes: Vec<BlokHeaderSubchunk> },
	Shdr { ordinal: Ordinal, attributes: Vec<BlokHeaderSubchunk> },
}
pub enum BlokHeaderSubchunk {
	Chan { texture_channel: u32 },
	Enab { enable: u16 },
	Opac { typecode: u16, opacity: f32, envelope: Vx },
	Axis { displacement_axis: u16 }
}
pub enum BlokSubchunk {
	Tmap { attributes: Vec<TmapSubchunk> },
	Proj { mode: u16 },
	Axis { texture_axis: u16 },
	Imag { texture: Vx },
	Wrap { width: u16, height: u16 },
	Wrpw { cycles: f32, envelope: Vx },
	Wrph { cycles: f32, envelope: Vx },
	Vmap { txuv_map_name: String },
	Aast { flags: u16, strength: f32 },
	Pixb { flags: u16 },
}
pub enum TmapSubchunk {
	Cntr { vector: Vector3f, envelope: Vx },
	Size { vector: Vector3f, envelope: Vx },
	Rota { vector: Vector3f, envelope: Vx },
	Oref { object_name: String },
	Fall { typecode: u16, vector: Vector3f, envelope: Vx },
	Csys { typecode: u16 },
}

///////////////////////////////////////
// Code

trait ChunkReader {
	fn read_chunk(&mut self) -> Result<Chunk>;
	fn read_subchunk(&mut self) -> Result<Subchunk>;
	fn read_vector3f(&mut self) -> Result<Vector3f>;
	fn read_string(&mut self) -> Result<String>;
	fn read_ordinal(&mut self) -> Result<Ordinal>;
	fn read_vx(&mut self) -> Result<u32>;
}
impl ChunkReader for Stream {
	fn read_chunk(&mut self) -> Result<Chunk> {
		let chunk = Chunk {
			tag: try!( self.read_u32() ),
			length: try!( self.read_u32() ),
		};

		Ok( chunk )
	}

	fn read_subchunk(&mut self) -> Result<Subchunk> {
		let subchunk = Subchunk {
			tag: try!( self.read_u32() ),
			length: try!( self.read_u16() ),
		};

		Ok( subchunk )
	}

	fn read_vector3f(&mut self) -> Result<Vector3f> {
		let v = Vector3f {
			v: [
				try!( self.read_f32() ),
				try!( self.read_f32() ),
				try!( self.read_f32() )
			]
		};

		Ok( v )
	}

	fn read_string(&mut self) -> Result<String> {
		let mut s = Vec::<u8>::new() ;
		let bytes = try!( self.read_until( '\0' as u8, &mut s ) );
		s.pop();		// delete '\0'
		if 1 == ( bytes & 0x1 ) {
			self.consume( 1 );
		}
		Ok( try!( String::from_utf8( s ) ) )
	}

	fn read_ordinal(&mut self) -> Result<Ordinal> {
		let mut s = Vec::<u8>::new() ;
		let bytes = try!( self.read_until( '\0' as u8, &mut s ) );
		if 1 == ( bytes & 0x1 ) {
			self.consume( 1 );
		}
		Ok( s )
	}

	fn read_vx(&mut self) -> Result<u32> {
		let val = try!( self.read_u16() );

		Ok( val as u32 )
	}
}

/*
trait ChunkReader {
	fn read_chunk(&mut self) -> Result<Chunk>;
	fn read_u32(&mut self) -> Result<u32>;
	fn read_u16(&mut self) -> Result<u16>;
	fn read_f32(&mut self) -> Result<f32>;
	fn read_vector3f(&mut self) -> Result<Vector3f>;
	fn read_string(&mut self) -> Result<String>;
}

impl ChunkReader for Read {
	fn read_chunk(&mut self) -> Result<Chunk> {
		let chunk = Chunk {
			tag: try!(self.read_u32()),
			length: try!(self.read_u32()),
		};

		Ok(chunk)
	}

	fn read_u32(&mut self) -> Result<u32> {
		let mut id: [u8; 4] = [0; 4];
		try!(self.read_exact(&mut id));
		unsafe {
			Ok(u32::from_be(mem::transmute::<[u8; 4], u32>(id)))
		}
	}

	fn read_u16(&mut self) -> Result<u16> {
		let mut id: [u8; 2] = [0; 2];
		try!(self.read_exact(&mut id));
		unsafe {
			Ok(u16::from_be(mem::transmute::<[u8; 2], u16>(id)))
		}
	}

	fn read_f32(&mut self) -> Result<f32> {
		unsafe {
			Ok(mem::transmute::<u32, f32>(try!(self.read_u32())))
		}
	}

	fn read_vector3f(&mut self) -> Result<Vector3f> {
		let v = Vector3f {
			v: [
				try!(self.read_f32()),
				try!(self.read_f32()),
				try!(self.read_f32())
			]
		};

		Ok(v)
	}

	fn read_until(&mut self, byte: u8, buf: &mut Vec<u8>) -> Result<usize> {

	}

	fn read_string(&mut self) -> Result<String> {
		let mut s = Vec::<u8>::new() ;
		let bytes = try!(self.read_until('\0' as u8, &mut s));
		if 1 == (bytes & 0x1) {
			self.consume(1);
		}
		Ok(try!(String::from_utf8(s)))
	}
}
*/

pub fn process(reader: &mut Stream) -> Result<Box<Object>> {
	let form = try!( reader.read_chunk() );
	match form.tag {
		id if id == ChunkIDs::FORM as u32 => handle_form( reader ),
		_ => Err( Box::new( LwoError::Invalid ) ),
	}
}

fn handle_form(reader: &mut Stream) -> Result<Box<Object>> {
	if try!( reader.read_u32() ) != Literals::LWO2 as u32 {
		return Err( Box::new( LwoError::Invalid ) )
	}

	let mut object = Box::new( Object {
		tags: Tags { v: Vec::new() },
		layers: Vec::new(),
		clips: Vec::new(),
		surfaces: Vec::new(),
	} );

	while reader.limit() != 0 {
		let child: Chunk = try!( reader.read_chunk() );
		let old_limit = reader.limit;
		reader.limit = child.length as usize;

		match child.tag {
			id if id == ChunkIDs::LAYR as u32 => {
				object.layers.push( try!( handle_layr( reader ) ) );
			},
			id if id == ChunkIDs::PNTS as u32 => {
				object.layers.last_mut().unwrap().points = try!( handle_pnts( reader ) );
			},
			id if id == ChunkIDs::VMAP as u32 => {
				object.layers.last_mut().unwrap().vmaps.push( try!( handle_vmap( reader ) ) );
			},
			id if id == ChunkIDs::VMAD as u32 => {
				object.layers.last_mut().unwrap().vmads.push( try!( handle_vmad( reader ) ) );
			},
			id if id == ChunkIDs::POLS as u32 => {
				object.layers.last_mut().unwrap().polygons = try!( handle_pols( reader ) );
			},
			id if id == ChunkIDs::TAGS as u32 => {
				object.tags = try!( handle_tags( reader ) );
			},
			id if id == ChunkIDs::PTAG as u32 => {
				object.layers.last_mut().unwrap().ptags = try!( handle_ptag( reader ) );
			},
			id if id == ChunkIDs::CLIP as u32 => {
				object.clips.push( try!( handle_clip( reader ) ) );
			},
			id if id == ChunkIDs::SURF as u32 => {
				object.surfaces.push( try!( handle_surf( reader ) ) );
			},
			
			_ => reader.consume( child.length as usize ),
		};
		reader.limit = old_limit - child.length as usize;
	}

	Ok( object )
}

fn handle_layr(reader: &mut Stream) -> Result<Box<Layr>> {	// TODO: lack of symmetry with rest of handlers
	 Ok( Box::new( Layr {
		number: try!( reader.read_u16() ),
		flags: try!( reader.read_u16() ),
		pivot: try!( reader.read_vector3f() ),
		name: try!( reader.read_string() ),
		parent: try!( reader.read_u16() ),	// TODO: this is optional!!!
		points: Box::new(Pnts { v: Vec::new() }),
		vmaps: Vec::new(),
		vmads: Vec::new(),
		polygons: Box::new( Pols { typecode: PolsType::Unknown( 0 ), v: Vec::new() } ),
		ptags: Box::new( Ptag { typecode: PtagType::Unknown( 0 ), v: Vec::new() } )
	}))
}

fn handle_pnts(reader: &mut Stream) -> Result<Box<Pnts>> {
	let mut pnts = Box::new( Pnts { v: Vec::new() } );
	while reader.limit() != 0  {
		pnts.v.push( try!( reader.read_vector3f() ) );
	}

	Ok( pnts )
}

fn handle_vmap(reader: &mut Stream) -> Result<Box<Vmap>> {
	let map_type = |t: u32| -> VmapType {
		match t {
			t if t == Literals::TXUV as u32 => VmapType::TXUV,
			_ => VmapType::Unknown( t )
		}
	};

	let mut vmap = Box::new( Vmap {
		typecode: map_type( try!( reader.read_u32() ) ),
		dimension: try!( reader.read_u16() ),
		name: try!( reader.read_string() ),
		v: Vec::new()
	});

	while reader.limit() != 0 {
		let vx = try!( reader.read_vx() );
		let mut v = Vec::new();
		for _ in 0..vmap.dimension {
			v.push( try!( reader.read_f32() ) );
		}
		vmap.v.push( VmapViVal { vi: vx, val: v } );
	}

	Ok(vmap)
}

fn handle_vmad(reader: &mut Stream) -> Result<Box<Vmad>> {
	let map_type = |t: u32| -> VmadType {
		match t {
			t if t == Literals::TXUV as u32 => VmadType::TXUV,
			_ => VmadType::Unknown( t )
		}
	};

	let mut vmad = Box::new( Vmad {
		typecode: map_type( try!( reader.read_u32() ) ),
		dimension: try!( reader.read_u16() ),
		name: try!( reader.read_string() ),
		v: Vec::new()
	});

	while reader.limit() != 0 {
		let vx = try!( reader.read_vx() );
		let px = try!( reader.read_vx() );
		let mut v = Vec::new();
		for _ in 0..vmad.dimension {
			v.push( try!( reader.read_f32() ) );
		}
		vmad.v.push( VmadViPiVal { vi: vx, pi: px, val: v } );
	}

	Ok( vmad )
}

fn handle_pols(reader: &mut Stream) -> Result<Box<Pols>> {
	let map_type = |t: u32| -> PolsType {
		match t {
			t if t == Literals::FACE as u32 => PolsType::FACE,
			_ => PolsType::Unknown( t )
		}
	};

	let mut pols = Box::new( Pols {
		typecode: map_type( try!( reader.read_u32() ) ),
		v: Vec::new(),
	});

	let mut flags;
	let mut n;
	while reader.limit() != 0 {
		n = try!( reader.read_u16() );		// TODO: mask
		flags = n & (0x3F << 10);		// TODO: mask
		n &= 0x3FF;
		let mut v = Vec::new();
		for _ in 0..n {
			v.push( try!( reader.read_vx() ) );
		}
		pols.v.push( (flags, v) );
	}

	Ok( pols )
}

fn handle_tags(reader: &mut Stream) -> Result<Tags> {
	let mut tags = Tags { v: Vec::new() };

	while reader.limit() != 0 {
		tags.v.push( try!( reader.read_string() ) );
	}

	Ok(tags)
}

fn handle_ptag(reader: &mut Stream) -> Result<Box<Ptag>> {
	let map_type = |t: u32| -> PtagType {
		match t {
			t if t == Literals::SURF as u32 => PtagType::SURF,
			_ => PtagType::Unknown( t )
		}
	};

	let mut ptag = Box::new( Ptag {
		typecode: map_type( try!( reader.read_u32() ) ),
		v: Vec::new(),
	});

	while reader.limit() != 0 {
		ptag.v.push( (try!( reader.read_u16() ) as u32, try!( reader.read_u16() )) );
	}

	Ok( ptag )
}

fn handle_clip(reader: &mut Stream) -> Result<Clip> {
	let mut clip = Clip {
		index: try!( reader.read_u32() ),
		vsub: Vec::new(),
	};


	while reader.limit() != 0 {
		let subchunk = try!( reader.read_subchunk() );
		let old_limit = reader.limit;
		reader.limit = subchunk.length as usize;

		match subchunk.tag {
			id if id == SubchunkIDs::STIL as u32 => {
				clip.vsub.push( try!( handle_sub_stil( reader) ) );
			},
			_ => reader.consume( subchunk.length as usize ),
		}
		reader.limit = old_limit - subchunk.length as usize;
	}

	Ok( clip )
}

fn handle_sub_stil(reader: &mut Stream) -> Result<ClipSubchunk> {
	Ok( ClipSubchunk::Stil {
		name: try!( reader.read_string() )
	} )
}

fn handle_surf(reader: &mut Stream) -> Result<Box<Surf>> {
	let mut surf = Box::new( Surf {
		name: try!( reader.read_string() ),
		source: try!( reader.read_string() ),
		attributes: Vec::new()
	} );


	while reader.limit() != 0 {
		let subchunk = try!(reader.read_subchunk());
		let old_limit = reader.limit;
		reader.limit = subchunk.length as usize;

		match subchunk.tag {
			id if id == SubchunkIDs::COLR as u32 => {
				surf.attributes.push(try!(handle_sub_colr(reader)));
			},
			id if id == SubchunkIDs::DIFF as u32 => {
				surf.attributes.push(try!(handle_sub_diff(reader)));
			},
			id if id == SubchunkIDs::SPEC as u32 => {
				surf.attributes.push(try!(handle_sub_spec(reader)));
			},
			id if id == SubchunkIDs::GLOS as u32 => {
				surf.attributes.push(try!(handle_sub_glos(reader)));
			},
			id if id == SubchunkIDs::BLOK as u32 => {
				surf.attributes.push(try!(handle_sub_blok(reader)));
			},
			_ => reader.consume(subchunk.length as usize),
		}
		reader.limit = old_limit - subchunk.length as usize;
	}

	Ok( surf )
}

fn handle_sub_colr(reader: &mut Stream) -> Result<SurfSubchunk> {
	Ok( SurfSubchunk::Colr {
		base: try!( reader.read_vector3f() ),
		envelope: try!( reader.read_vx() )
	} )
}

fn handle_sub_diff(reader: &mut Stream) -> Result<SurfSubchunk> {
	Ok( SurfSubchunk::Diff {
		intensity: try!( reader.read_f32() ),
		envelope: try!( reader.read_vx() )
	} )
}

fn handle_sub_spec(reader: &mut Stream) -> Result<SurfSubchunk> {
	Ok( SurfSubchunk::Spec {
		intensity: try!( reader.read_f32() ),
		envelope: try!( reader.read_vx() )
	} )
}

fn handle_sub_glos(reader: &mut Stream) -> Result<SurfSubchunk> {
	Ok( SurfSubchunk::Glos {
		intensity: try!( reader.read_f32() ),
		envelope: try!( reader.read_vx() )
	} )
}

fn handle_sub_blok(reader: &mut Stream) -> Result<SurfSubchunk> {
	let subchunk = try!( reader.read_subchunk() );
	let old_limit = reader.limit;
	reader.limit = subchunk.length as usize;

	let mut blok = SurfSubchunk::Blok {
		header: match subchunk.tag {
			id if id == SubchunkIDs::IMAP as u32 => {
				try!( handle_sub_imap( reader ) )
			},
			_ => return Err( Box::new( LwoError::UnknownBlokHeader ) )
		},
		attributes: Vec::new()
	};
	reader.limit = old_limit - subchunk.length as usize;
	
	while reader.limit() != 0 {
		let subchunk = try!( reader.read_subchunk() );
		let old_limit = reader.limit;
		reader.limit = subchunk.length as usize;

		match subchunk.tag {
			id if id == SubchunkIDs::TMAP as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_tmap( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::PROJ as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_proj( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::AXIS as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_axis( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::IMAG as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_imag( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::WRAP as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_wrap( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::WRPW as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_wrpw( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::WRPH as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_wrph( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::VMAP as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_vmap( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::AAST as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_aast( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::PIXB as u32 => {
				match &mut blok {
					&mut SurfSubchunk::Blok { header: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_pixb( reader ) ) ),
					_ => (),
				};
			},
			_ => reader.consume( subchunk.length as usize ),
		}
		reader.limit = old_limit - subchunk.length as usize;
	}

	Ok( blok )
}

fn handle_sub_imap(reader: &mut Stream) -> Result<BlokHeader> {
	let mut imap = BlokHeader::Imap {
		ordinal: try!( reader.read_ordinal() ),
		attributes: Vec::new()
	};
	
	while reader.limit() != 0 {
		let subchunk = try!( reader.read_subchunk() );
		let old_limit = reader.limit;
		reader.limit = subchunk.length as usize;

		match subchunk.tag {
			id if id == SubchunkIDs::CHAN as u32 => {
				match &mut imap {
					&mut BlokHeader::Imap { ordinal: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_chan( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::ENAB as u32 => {
				match &mut imap {
					&mut BlokHeader::Imap { ordinal: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_enab( reader) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::OPAC as u32 => {
				match &mut imap {
					&mut BlokHeader::Imap { ordinal: _, attributes: ref mut attr } =>
						attr.push( try!( handle_sub_opac( reader ) ) ),
					_ => (),
				};
			},
			_ => reader.consume( subchunk.length as usize ),
		}
		reader.limit = old_limit - subchunk.length as usize;
	}

	Ok( imap )
}

fn handle_sub_chan(reader: &mut Stream) -> Result<BlokHeaderSubchunk> {
	let chan = BlokHeaderSubchunk::Chan {
		texture_channel: try!( reader.read_u32() ),
	};

	Ok( chan )
}

fn handle_sub_enab(reader: &mut Stream) -> Result<BlokHeaderSubchunk> {
	let enab = BlokHeaderSubchunk::Enab {
		enable: try!( reader.read_u16() ),
	};

	Ok( enab )
}

fn handle_sub_opac(reader: &mut Stream) -> Result<BlokHeaderSubchunk> {
	let opac = BlokHeaderSubchunk::Opac {
		typecode: try!( reader.read_u16() ),
		opacity: try!( reader.read_f32() ),
		envelope: try!( reader.read_vx() ),
	};

	Ok( opac )
}
/*
fn handle_sub_axis(reader: &mut Stream) -> Result<BlokHeaderSubchunk> {
	let axis = BlokHeaderSubchunk::Axis {
		displacement_axis: try!(reader.read_u16()),
	};

	Ok(axis)
}
*/
fn handle_sub_tmap(reader: &mut Stream) -> Result<BlokSubchunk> {
	let mut tmap = BlokSubchunk::Tmap {
		attributes: Vec::new()
	};

	while reader.limit() != 0 {
		let subchunk = try!( reader.read_subchunk() );
		let old_limit = reader.limit;
		reader.limit = subchunk.length as usize;

		match subchunk.tag {
			id if id == SubchunkIDs::CNTR as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( try!( handle_sub_cntr( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::SIZE as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( try!( handle_sub_size( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::ROTA as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( try!( handle_sub_rota( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::OREF as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( try!( handle_sub_oref( reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::FALL as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( try!( handle_sub_fall(reader ) ) ),
					_ => (),
				};
			},
			id if id == SubchunkIDs::CSYS as u32 => {
				match &mut tmap {
					&mut BlokSubchunk::Tmap { attributes: ref mut attr } =>
						attr.push( try!( handle_sub_csys( reader ) ) ),
					_ => (),
				};
			},
			_ => reader.consume( subchunk.length as usize ),
		}
		reader.limit = old_limit - subchunk.length as usize;
	}

	Ok( tmap )
}

fn handle_sub_cntr(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Cntr {
		vector: try!( reader.read_vector3f() ),
		envelope: try!( reader.read_vx() ),
	} )
}

fn handle_sub_size(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Size {
		vector: try!( reader.read_vector3f() ),
		envelope: try!( reader.read_vx() ),
	} )
}

fn handle_sub_rota(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Rota {
		vector: try!( reader.read_vector3f() ),
		envelope: try!( reader.read_vx() ),
	} )
}

fn handle_sub_oref(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Oref {
		object_name: try!( reader.read_string() ),
	} )
}

fn handle_sub_fall(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Fall {
		typecode: try!( reader.read_u16() ),
		vector: try!( reader.read_vector3f() ),
		envelope: try!( reader.read_vx() ),
	} )
}

fn handle_sub_csys(reader: &mut Stream) -> Result<TmapSubchunk> {
	Ok( TmapSubchunk::Csys {
		typecode: try!( reader.read_u16() ),
	} )
}

fn handle_sub_proj(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Proj {
		mode: try!( reader.read_u16() ),
	} )
}

fn handle_sub_axis(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Axis {
		texture_axis: try!( reader.read_u16() ),
	} )
}

fn handle_sub_imag(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Imag {
		texture: try!( reader.read_vx() ),
	} )
}

fn handle_sub_wrap(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Wrap {
		width: try!( reader.read_u16() ),
		height: try!( reader.read_u16() ),
	} )
}

fn handle_sub_wrpw(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Wrpw {
		cycles: try!( reader.read_f32() ),
		envelope: try!( reader.read_vx() ),
	} )
}

fn handle_sub_wrph(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Wrpw {
		cycles: try!( reader.read_f32() ),
		envelope: try!( reader.read_vx() ),
	} )
}

fn handle_sub_vmap(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Vmap {
		txuv_map_name: try!( reader.read_string() ),
	} )
}

fn handle_sub_aast(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Aast {
		flags: try!( reader.read_u16() ),
		strength: try!( reader.read_f32() ),
	} )
}

fn handle_sub_pixb(reader: &mut Stream) -> Result<BlokSubchunk> {
	Ok( BlokSubchunk::Pixb {
		flags: try!( reader.read_u16() ),
	} )
}
