use std::io;
use vector::*;

///////////////////////////////////////
// Mesh description

pub struct Vertex {
	pub pos: Vector3<i16>,
	pub tc: Vector2<i16>,
	pub col: Vector4<u8>
}

pub enum Face {
	Tri { ids: Vector3<i16> },
	Quad { ids: Vector4<i16> },
}

pub struct Texture {
	pub width: u16,
	pub height: u16,
	pub data: TextureData,
}
pub enum TextureData {
	CI4 { pixels: Vec<u8>, palette: Vec<u16> },
	RGBA16 { pixels: Vec<u16> },
	Unknown { pixels: Vec<u8> }
}

pub struct Material {
	pub tex_id: Option<usize>,
	pub color: Vector3<f32>,
}

pub struct Layer {
	pub verts: Vec<Vertex>,
	pub faces: Vec<Face>,
	pub face2mtl: Vec<(usize, usize)>
}

pub struct Mesh {
	pub layers: Vec<Layer>,
	pub materials: Vec<Material>,
	pub textures: Vec<Texture>
}

///////////////////////////////////////
// Code

pub fn save(mesh: &Mesh, buffer: &mut io::Write) -> Result<(), io::Error> {
	let header = "#include \"myheader.h\"\n";

	try!( buffer.write_all( header.as_bytes() ) );
	try!( buffer.write_all( b"\n") );

	for (i,layer) in mesh.layers.iter().enumerate() {
		try!( save_vts( &(i,layer), buffer ) );
		try!( save_faces( &(i,layer), buffer ) );
	}

	try!( save_textures( mesh, buffer ) );
	try!( save_texture_attributes( mesh, buffer) );

	// materials
	try!( write!( buffer, "static Material materials[] = {{\n" ) );
	for material in &mesh.materials {
		match material.tex_id {
			Some(tid) => {
				try!( write!( buffer, "\t{{ {{ {}, {}, {} }}, &textures[{}] }},\n",
					material.color.v[0], material.color.v[1], material.color.v[2],
					tid
				) );
			},
			None => {
				try!( write!( buffer, "\t{{ {{ {}, {}, {} }}, NULL }},\n",
					material.color.v[0], material.color.v[1], material.color.v[2]
				) );
			}
		}
	}
	try!( close_struct( buffer ) );

	// pairs
	try!( write!( buffer, "Pair tab[] = {{\n" ) );
	for (i,_) in mesh.layers.iter().enumerate() {
		try!( write!( buffer, "\t{{ &materials[{}], drawmesh{} }},\n", i, i ) );
	}
	try!( write!( buffer, "\t{{ NULL, NULL }}\n" ) );
	try!( close_struct( buffer ) );

	Ok( () )
}

fn create_texture_name(texid: usize) -> String {
	String::from( "texture" ) + texid.to_string().as_str() 
}

fn close_struct(buffer: &mut io::Write) -> Result<(), io::Error> {
	try!( buffer.write_all( b"};\n\n" ) );

	Ok( () )
}

fn save_vts(&(i,layer): &(usize,&Layer), buffer: &mut io::Write) -> Result<(), io::Error> {
	try!( write!( buffer, "static Vtx vts{}[] = {{\n", i ) );

	for vtx in layer.verts.iter() {
		try!( write!( buffer, "\t{{\n" ) );
		try!( write!( buffer, "\t\t{{ {{ {}, {}, {} }}, 0, {{ {}, {} }}, {{ 0x{:02x}, 0x{:02x}, 0x{:02x}, 0x{:02x} }} }}\n",
			vtx.pos.v[0], vtx.pos.v[1], vtx.pos.v[2],
			vtx.tc.v[0], vtx.tc.v[1],
			vtx.col.v[0], vtx.col.v[1], vtx.col.v[2], vtx.col.v[3] ) );
		try!( write!( buffer, "\t}},\n" ) );
	}

	try!( close_struct( buffer ) );

	Ok( () )
}

fn save_faces(&(i,layer): &(usize,&Layer), buffer: &mut io::Write) -> Result<(), io::Error> {
	try!( write!( buffer, "static Gfx drawmesh{}[] = {{\n", i ) );

	for face in &layer.faces {
		match face {
			&Face::Tri { ref ids } => {
				try!( write!( buffer, "\tgsSPVertex(&vts{}[{}], 1, 0),\n", i, ids.v[0] ) );
				try!( write!( buffer, "\tgsSPVertex(&vts{}[{}], 1, 1),\n", i, ids.v[1] ) );
				try!( write!( buffer, "\tgsSPVertex(&vts{}[{}], 1, 2),\n", i, ids.v[2] ) );
				try!( write!( buffer, "\tgsSP1Triangle(0, 1, 2, 0),\n"));
			},
			&Face::Quad { ref ids } => {
				try!( write!( buffer, "\tgsSPVertex(&vts{}[{}], 1, 0),\n", i, ids.v[0] ) );
				try!( write!( buffer, "\tgsSPVertex(&vts{}[{}], 1, 1),\n", i, ids.v[1] ) );
				try!( write!( buffer, "\tgsSPVertex(&vts{}[{}], 1, 2),\n", i, ids.v[2] ) );
				try!( write!( buffer, "\tgsSPVertex(&vts{}[{}], 1, 3),\n", i, ids.v[3] ) );
				try!( write!( buffer, "\tgsSP2Triangles(0, 1, 2, 0, 0, 2, 3, 0),\n" ) );
			},
		}
	}

	try!( buffer.write_all( b"\tgsSPEndDisplayList()\n" ) );
	try!( close_struct( buffer ) );

	Ok( () )
}

fn save_textures(mesh: &Mesh, buffer: &mut io::Write) -> Result<(), io::Error> {
	let write_pal = |buffer: &mut io::Write, palette: &Vec<u16>, i: usize| -> Result<(), io::Error> {
		try!( write!( buffer, "static u16 palette{}[] = {{\n\t", i ) );

		for (i,color) in palette.iter().enumerate() {
			if ( i != 0 ) && ( i%8 == 0 ) {
				try!( write!( buffer, "\n\t" ) );
			}
			try!( write!( buffer, "0x{:04x}, ", color ) );
		}

		try!( write!( buffer, "\n" ) );
		try!( close_struct( buffer ) );

		Ok( () )
	};

	for (i,texture) in mesh.textures.iter().enumerate() {
		match &texture.data {
			&TextureData::CI4 { ref pixels, ref palette } => {
				try!( write_pal( buffer, palette, i ) );
				try!( write!( buffer, "static u8 {}[] = {{\n\t", create_texture_name( i ) ) );

				for (i,color) in pixels.iter().enumerate() {
					if ( i != 0 ) && ( i&3 == 0 ) {
						try!( write!( buffer, "\n\t" ) );
					}
					try!( write!( buffer, "0x{:02x}, ", color ) );
				}

				if pixels.is_empty() {
					try!( write!( buffer, "\t0\t//dummy element - texture is empty" ) );
				}
			},
			&TextureData::RGBA16 { ref pixels } => {
				try!( write!( buffer, "static u16 {}[] = {{\n\t", create_texture_name( i ) ) );

				for (i,color) in pixels.iter().enumerate() {
					if ( i != 0 ) && ( i&3 == 0 ) {
						try!( write!( buffer, "\n\t" ) );
					}
					try!( write!( buffer, "0x{:04x}, ", color ) );
				}

				if pixels.is_empty() {
					try!( write!( buffer, "\t0\t//dummy element - texture is empty" ) );
				}
			},
			&TextureData::Unknown { ref pixels } => {
				try!( write!( buffer, "static u32 {}[] = {{\n\t", create_texture_name( i ) ) );

				for (i,color) in pixels.iter().enumerate() {
					if ( i != 0 ) && ( i&3 == 0 ) {
						try!( write!( buffer, "\n\t" ) );
					}
					try!( write!( buffer, "0x{:02x}, ", color ) );
				}

				if pixels.is_empty() {
					try!( write!( buffer, "\t0\t//dummy element - texture is empty" ) );
				}
			}
		};

		try!( write!( buffer, "\n" ) );
		try!( close_struct( buffer ) );
	}

	Ok( () )
}

fn save_texture_attributes(mesh: &Mesh, buffer: &mut io::Write) -> Result<(), io::Error> {
	if mesh.textures.is_empty() {
		return Ok( () )
	}

	try!( write!( buffer, "static Texture textures[] = {{\n" ) );

	for (i,texture) in mesh.textures.iter().enumerate() {
		let (fmt_string, pal_string) = match texture.data {
			TextureData::CI4 { .. } => ("TEXTURE_FMT_CI4", String::from( "&palette" ) + i.to_string().as_str()),
			TextureData::RGBA16 { .. } => ("TEXTURE_FMT_RGBA16", String::from( "NULL" )),
			TextureData::Unknown { .. } => ("TEXTURE_FMT_UNKNOWN", String::from( "NULL" )),
		};
		try!(write!(buffer, "\tTEXTURE_INIT({}, {}, &texture{}, {}, {}),\n", texture.width, texture.height, i, pal_string, fmt_string)); 
	}

	try!( close_struct( buffer ) );

	Ok( () )
}
